from typing import Type, TYPE_CHECKING, TypeVar, Generic
from django.db import models

if TYPE_CHECKING:
    from django.db.models.manager import ManyToManyRelatedManager


T = TypeVar('T')
M = TypeVar('M', bound=models.Model)


def with_typehint(baseclass: Type[T]) -> Type[T]:
    """
    Useful function to make mixins with baseclass typehint

    ```
    class ReadonlyMixin(with_typehint(BaseAdmin))):
        ...
    ```
    """
    if TYPE_CHECKING:
        return baseclass

    return object  # noqa


if not TYPE_CHECKING:
    class ManyToManyRelatedManager(Generic[M]):
        """
        Pydantic Compatible Generic Type for declaring a ManyToManyField
        """
        @classmethod
        def __get_validators__(cls):
            yield lambda v: v



from decimal import Decimal
import os
import json
from typing import Coroutine, Mapping, Optional, Type, Union
from contextvars import ContextVar
from django.db.models.query_utils import DeferredAttribute
from pydantic import BaseModel, parse_obj_as
from pydantic.fields import ModelField, SHAPE_SINGLETON, SHAPE_LIST, Undefined
from pydantic.types import ConstrainedStr
from django.db import models
from django.db.models.manager import Manager
from django.db.models.fields.related_descriptors import ManyToManyDescriptor, ReverseManyToOneDescriptor
from django.utils.functional import cached_property
from sentry_tools.decorators import instrument_span
from sentry_tools.span import set_tag, set_data
from async_tools import is_async, sync_to_async
from ...schemas import AccessScope
from ...exceptions import AccessError
from ... import context
from ...fields import ORMFieldInfo


transfer_current_obj: ContextVar[models.Model] = ContextVar('transfer_current_obj')


class Break(Exception):
    """
    Internal Exception used to break lookping through all object attributes and instead use the given value
    """


@instrument_span(
    op='transfer_from_orm',
    description=lambda pydantic_cls, django_obj, *args, **kwargs: f'{django_obj} to {pydantic_cls.__name__}',
)
def transfer_from_orm(
    pydantic_cls: Type[BaseModel],
    django_obj: models.Model,
    django_parent_obj: Optional[models.Model] = None,
    pydantic_field_on_parent: Optional[ModelField] = None,
    filter_submodel: Optional[Mapping[Manager, models.Q]] = None,
) -> Union[BaseModel, Coroutine[None, None, BaseModel]]:
    """
    Transfers the field contents of django_obj to a new instance of pydantic_cls.
    For this to work it is required to have orm_field set on all of the pydantic_obj's fields, which has to point to the django model attribute.

    It also works for nested pydantic models which point to a field on the **same** django model and for related fields (m2o or m2m).

    Example:

    ```python
    from pydantic import BaseModel, Field
    from django.db import models

    class Address(models.Model):
        name = models.CharField(max_length=56)

    class AddressRequest(BaseModel):
        name: str = Field(orm_field=Address.name)
    ```
    """
    if is_async():
        return sync_to_async(_transfer_from_orm)(
            pydantic_cls=pydantic_cls,
            django_obj=django_obj,
            django_parent_obj=django_parent_obj,
            pydantic_field_on_parent=pydantic_field_on_parent,
            filter_submodel=filter_submodel,
        )

    return _transfer_from_orm(
        pydantic_cls=pydantic_cls,
        django_obj=django_obj,
        django_parent_obj=django_parent_obj,
        pydantic_field_on_parent=pydantic_field_on_parent,
        filter_submodel=filter_submodel,
    )


def _compute_value_from_orm_method(
    orm_method: callable,
    field: ModelField,
    django_obj: models.Model,
    filter_submodel: Optional[Mapping[Manager, models.Q]] = None,
):
    value = orm_method(django_obj)
    if value is not None and issubclass(field.type_, BaseModel) and not isinstance(value, BaseModel):
        if field.shape == SHAPE_SINGLETON:
            if isinstance(value, models.Model):
                value = _transfer_from_orm(
                    pydantic_cls=field.type_,
                    django_obj=value,
                )

            else:
                value = field.type_.parse_obj(value)

        elif field.shape == SHAPE_LIST:
            def _to_pydantic(obj):
                if isinstance(obj, BaseModel):
                    return obj

                if isinstance(obj, models.Model):
                    return _transfer_from_orm(
                        pydantic_cls=field.type_,
                        django_obj=obj,
                        django_parent_obj=django_obj,
                        filter_submodel=filter_submodel,
                    )

                return field.type_.parse_obj(obj)

            value = [
                _to_pydantic(obj)
                for obj in value
            ]

        else:
            raise NotImplementedError

    return value


def _transfer_field_list(
    field: ModelField,
    orm_field,
    django_obj: models.Model,
    filter_submodel: Optional[Mapping[Manager, models.Q]] = None,
):
    sub_filter = filter_submodel and filter_submodel.get(orm_field) or models.Q()

    if isinstance(orm_field, ManyToManyDescriptor):
        relatedmanager = getattr(django_obj, orm_field.field.attname)
        related_objs = relatedmanager.through.objects.filter(models.Q(**{relatedmanager.source_field_name: relatedmanager.instance}) & sub_filter)

    elif isinstance(orm_field, ReverseManyToOneDescriptor):
        relatedmanager = getattr(django_obj, orm_field.rel.name)
        related_objs = relatedmanager.filter(sub_filter)

    elif isinstance(orm_field, DeferredAttribute) and isinstance(orm_field.field, models.JSONField):
        value = None
        try:
            value = getattr(django_obj, orm_field.field.attname)

        except AttributeError:
            raise  # attach debugger here ;)

        return parse_obj_as(field.outer_type_, value or [])

    else:
        raise NotImplementedError

    return [
        _transfer_from_orm(
            pydantic_cls=field.type_,
            django_obj=rel_obj,
            django_parent_obj=django_obj,
            pydantic_field_on_parent=field,
            filter_submodel=filter_submodel,
        ) for rel_obj in related_objs
    ]


def _transfer_field_singleton(
    field: ModelField,
    orm_field,
    django_obj: models.Model,
    pydantic_field_on_parent: Optional[ModelField] = None,
    filter_submodel: Optional[Mapping[Manager, models.Q]] = None,
):
    if not orm_field and issubclass(field.type_, BaseModel):
        return _transfer_from_orm(
            pydantic_cls=field.type_,
            django_obj=django_obj,
            pydantic_field_on_parent=field,
            filter_submodel=filter_submodel,
        )

    value = None
    is_property = isinstance(orm_field, (property, cached_property))
    is_django_field = not is_property

    try:
        if is_property:
            if isinstance(orm_field, property):
                value = orm_field.fget(django_obj)

            elif isinstance(orm_field, cached_property):
                value = orm_field.__get__(django_obj)

            else:
                raise NotImplementedError

            if isinstance(value, models.Model):
                value = value.pk

        else:
            value = getattr(django_obj, orm_field.field.attname)

    except AttributeError:
        raise  # attach debugger here ;)

    if field.required and pydantic_field_on_parent and pydantic_field_on_parent.allow_none and value is None:
        raise Break(None)

    if is_django_field and value and isinstance(orm_field.field, models.JSONField):
        if issubclass(field.type_, BaseModel):
            if isinstance(value, dict):
                value = field.type_.parse_obj(value)

            else:
                value = field.type_.parse_raw(value)

        elif issubclass(field.type_, dict):
            if isinstance(value, str):
                value = json.loads(value)

        else:
            raise NotImplementedError

    scopes = [AccessScope.from_str(audience) for audience in (field.field_info.scopes if isinstance(field.field_info, ORMFieldInfo) else field.field_info.extra.get('scopes')) or []]
    if scopes:
        try:
            access = context.access.get()

        except LookupError:
            pass

        else:
            read_scopes = [str(scope) for scope in scopes if scope.action == 'read']
            if read_scopes:
                _value = None
                if not field.allow_none:
                    if issubclass(field.type_, str):
                        _value = '•' * (field.type_.max_length if issubclass(field.type_, ConstrainedStr) else len(value))

                    elif issubclass(field.type_, (int, float, Decimal)):
                        _value = 0

                    else:
                        raise NotImplementedError

                if not access.token.has_audience(read_scopes):
                    value = _value

                else:
                    if hasattr(django_obj, 'check_access'):
                        for scope in scopes:
                            if scope.action != 'read':
                                continue

                            try:
                                django_obj.check_access(access, selector=scope.selector)

                            except AccessError:
                                value = _value

    return value


def _transfer_field(
    field: ModelField,
    django_obj: models.Model,
    pydantic_field_on_parent: Optional[ModelField] = None,
    filter_submodel: Optional[Mapping[Manager, models.Q]] = None,
):
    orm_method = field.field_info.orm_method if isinstance(field.field_info, ORMFieldInfo) else field.field_info.extra.get('orm_method')
    if orm_method:
        return _compute_value_from_orm_method(
            orm_method=orm_method,
            field=field,
            django_obj=django_obj,
            filter_submodel=filter_submodel,
        )

    if isinstance(field.field_info, ORMFieldInfo):
        orm_field = field.field_info.orm_field
        if orm_field is Undefined:
            return ...

    else:
        orm_field = field.field_info.extra.get('orm_field')
        if 'orm_field' in field.field_info.extra and field.field_info.extra['orm_field'] is None:
            # Do not raise error when orm_field was explicitly set to None
            return ...

    if not orm_field and not (field.shape == SHAPE_SINGLETON and issubclass(field.type_, BaseModel)):
        raise AttributeError("orm_field not found on %r (parent: %r)" % (field, pydantic_field_on_parent))

    if field.shape == SHAPE_SINGLETON:
        return _transfer_field_singleton(
            field=field,
            orm_field=orm_field,
            django_obj=django_obj,
            filter_submodel=filter_submodel,
            pydantic_field_on_parent=pydantic_field_on_parent,
        )

    if field.shape == SHAPE_LIST:
        return _transfer_field_list(
            field=field,
            orm_field=orm_field,
            django_obj=django_obj,
            filter_submodel=filter_submodel,
        )

    raise NotImplementedError


def _transfer_from_orm(
    pydantic_cls: Type[BaseModel],
    django_obj: models.Model,
    django_parent_obj: Optional[models.Model] = None,
    pydantic_field_on_parent: Optional[ModelField] = None,
    filter_submodel: Optional[Mapping[Manager, models.Q]] = None,
) -> Union[BaseModel, Coroutine[None, None, BaseModel]]:
    set_tag('transfer_from_orm.pydantic_cls', pydantic_cls.__name__)
    set_tag('transfer_from_orm.django_cls', django_obj.__class__.__name__)
    set_data('transfer_from_orm.django_obj', django_obj)
    set_data('transfer_from_orm.django_parent_obj', django_parent_obj)
    set_data('transfer_from_orm.filter_submodel', filter_submodel)

    transfer_current_obj.set(django_obj)

    values = {}
    field: ModelField
    for field in pydantic_cls.__fields__.values():
        try:
            value = _transfer_field(
                field=field,
                django_obj=django_obj,
                pydantic_field_on_parent=pydantic_field_on_parent,
                filter_submodel=filter_submodel,
            )

        except Break as break_:
            # The whole object should be None
            return break_.args[0]

        else:
            if value is ...:
                continue

            values[field.name] = value

    return pydantic_cls.construct(**values)



import warnings
from typing import Callable, List, Optional, Tuple, Type
from enum import Enum
from pydantic import BaseModel, validate_model, SecretStr
from pydantic.fields import SHAPE_SINGLETON, SHAPE_LIST, Undefined
from django.db import models
from django.db.models.fields import Field as DjangoField
from django.db.models.fields.related_descriptors import ManyToManyDescriptor, ReverseManyToOneDescriptor
from django.db.transaction import atomic
from sentry_tools.decorators import instrument_span
from sentry_tools.span import set_tag, set_data
from async_tools import is_async, sync_to_async
from ...schemas import Access
from ..pydantic import Reference, get_orm_field_attr, is_orm_field_set
from .checks import check_field_access

try:
    from dirtyfields import DirtyFieldsMixin

except ImportError:
    class DirtyFieldsMixin:
        pass

try:
    from fastapi.exceptions import RequestValidationError

except ImportError:
    from pydantic import ValidationError as RequestValidationError


class TransferAction(Enum):
    CREATE = 'CREATE'
    SYNC = 'SYNC'
    NO_SUBOBJECTS = 'NO_SUBOBJECTS'


@instrument_span(
    op='transfer_to_orm',
    description=lambda pydantic_obj, django_obj, *args, **kwargs: f'{pydantic_obj} to {django_obj}',
)
def transfer_to_orm(
    pydantic_obj: BaseModel,
    django_obj: models.Model,
    *,
    action: Optional[TransferAction] = None,
    exclude_unset: bool = False,
    access: Optional[Access] = None,
    created_submodels: Optional[List[models.Model]] = None,
    _just_return_objs: bool = False,
    do_not_save_if_no_change: bool = False,
) -> Optional[Tuple[List[models.Model], List[models.Model]]]:
    """
    Transfers the field contents of pydantic_obj to django_obj.
    For this to work it is required to have orm_field set on all of the pydantic_obj's fields, which has to point to the django model attribute.

    It also works for nested pydantic models which point to a field on the **same** django model.

    Example:

    ```python
    from pydantic import BaseModel, Field
    from django.db import models

    class Address(models.Model):
        name = models.CharField(max_length=56)

    class AddressRequest(BaseModel):
        name: str = Field(orm_field=Address.name)
    ```
    """
    if is_async():
        return sync_to_async(transfer_to_orm)(
            pydantic_obj=pydantic_obj,
            django_obj=django_obj,
            action=action,
            exclude_unset=exclude_unset,
            access=access,
            created_submodels=created_submodels,
        )

    set_tag('transfer_to_orm.action', action)
    set_tag('transfer_to_orm.exclude_unset', exclude_unset)
    set_data('transfer_to_orm.access', access)
    set_data('transfer_to_orm.pydantic_obj', pydantic_obj)
    set_data('transfer_to_orm.django_obj', django_obj)

    if created_submodels:
        warnings.warn("Use transfer_to_orm with kwarg action instead of created_submodels", category=DeprecationWarning)
        action = TransferAction.CREATE

    if not action:
        warnings.warn("Use transfer_to_orm with kwarg action", category=DeprecationWarning)

    subobjects: List[models.Model] = created_submodels or []
    existing_objects = []

    if access:
        check_field_access(pydantic_obj, access)

    pydantic_values: Optional[dict] = pydantic_obj.dict(exclude_unset=True) if exclude_unset else None

    def populate_default(pydantic_cls: BaseModel, django_obj):
        for key, field in pydantic_cls.__fields__.items():
            orm_field = get_orm_field_attr(field.field_info, 'orm_field')
            if not orm_field and issubclass(field.type_, BaseModel):
                populate_default(field.type_, django_obj)

            else:
                if not is_orm_field_set(field.field_info):
                    continue

                if get_orm_field_attr(field.field_info, 'orm_method'):
                    # Do not raise error when orm_method is set
                    continue

                orm_field: DjangoField = get_orm_field_attr(field.field_info, 'orm_field')

                assert orm_field, "orm_field not set on %r of %r" % (field, pydantic_cls)

                setattr(
                    django_obj,
                    orm_field.field.attname,
                    field.field_info.default if field.field_info.default is not Undefined and field.field_info.default is not ... else None,
                )

    for key, field in pydantic_obj.__fields__.items():
        orm_method = get_orm_field_attr(field.field_info, 'orm_method')
        if orm_method:
            if exclude_unset and key not in pydantic_values:
                continue

            value = getattr(pydantic_obj, field.name)
            if isinstance(value, SecretStr):
                value = value.get_secret_value()

            orm_method(django_obj, value)
            continue

        if not is_orm_field_set(field.field_info):
            continue

        orm_field = get_orm_field_attr(field.field_info, 'orm_field')
        if not orm_field and not (field.shape == SHAPE_SINGLETON and issubclass(field.type_, BaseModel)):
            raise AttributeError("orm_field not found on %r" % field)

        value = getattr(pydantic_obj, field.name)
        if field.shape == SHAPE_SINGLETON:
            if not orm_field and issubclass(field.type_, BaseModel):
                if value is None:
                    if exclude_unset and key not in pydantic_values:
                        continue

                    populate_default(field.type_, django_obj)

                elif isinstance(value, BaseModel):
                    sub_transfer = transfer_to_orm(pydantic_obj=value, django_obj=django_obj, exclude_unset=exclude_unset, access=access, action=action, _just_return_objs=True)
                    subobjects += sub_transfer[0]
                    existing_objects += sub_transfer[1]

                else:
                    raise NotImplementedError

            else:
                if exclude_unset and key not in pydantic_values:
                    continue

                if orm_field.field.is_relation and isinstance(value, models.Model):
                    value = value.pk

                if isinstance(orm_field.field, models.JSONField) and value:
                    if isinstance(value, BaseModel):
                        value = value.dict()

                    elif isinstance(value, dict):
                        pass

                    else:
                        raise NotImplementedError

                setattr(django_obj, orm_field.field.attname, value)

        elif field.shape == SHAPE_LIST:
            if not value:
                continue

            elif isinstance(orm_field, ManyToManyDescriptor):
                relatedmanager = getattr(django_obj, orm_field.field.attname)
                related_model = relatedmanager.through
                obj_fields = {relatedmanager.source_field_name: django_obj}
                existing_object_ids = set()
                if action == TransferAction.SYNC:
                    existing_object_ids = set(related_model.objects.filter(**obj_fields).values_list('id', flat=True))

                for val in value:
                    def get_subobj(force_create: bool = False):
                        obj_manytomany_fields = {**obj_fields}
                        if getattr(val, 'id', None):
                            obj_manytomany_fields[relatedmanager.target_field.attname] = val.id

                        else:
                            raise NotImplementedError

                        if force_create or action == TransferAction.CREATE:
                            return related_model(**obj_manytomany_fields)

                        elif action == TransferAction.SYNC:
                            try:
                                return related_model.objects.get(**obj_manytomany_fields)

                            except related_model.DoesNotExist:
                                return get_subobj(force_create=True)

                        else:
                            raise NotImplementedError

                    sub_obj = get_subobj()
                    existing_object_ids.discard(sub_obj.id)
                    subobjects.append(sub_obj)
                    sub_transfer = transfer_to_orm(val, sub_obj, exclude_unset=exclude_unset, access=access, action=action, _just_return_objs=True)
                    subobjects += sub_transfer[0]
                    existing_objects += sub_transfer[1]

                existing_objects += related_model.objects.filter(id__in=list(existing_object_ids))

            elif isinstance(orm_field, ReverseManyToOneDescriptor):
                relatedmanager = getattr(django_obj, orm_field.rel.name)
                related_model: Type[models.Model] = relatedmanager.field.model
                obj_fields = {relatedmanager.field.name: django_obj}

                existing_object_ids = set()
                if action == TransferAction.SYNC:
                    existing_object_ids = set(related_model.objects.filter(**obj_fields).values_list('id', flat=True))

                val: BaseModel
                for val in value:
                    def get_subobj(force_create: bool = False):
                        matching = get_orm_field_attr(field.field_info, 'sync_matching')  # TODO replace with get_sync_matching_filter(val)
                        if action == TransferAction.SYNC and not getattr(val, 'id', None) and not matching:
                            force_create = True

                        if force_create or action == TransferAction.CREATE:
                            create_fields = {}
                            if getattr(val, 'id', None):
                                create_fields['id'] = getattr(val, 'id', None)

                            return related_model(**obj_fields, **create_fields)

                        elif action == TransferAction.SYNC:
                            try:
                                if getattr(val, 'id', None):
                                    return related_model.objects.get(id=val.id, **obj_fields)

                                elif matching:
                                    if isinstance(matching, list):
                                        matching_search = models.Q()
                                        pydantic_field_name: str
                                        match_orm_field: models.Field
                                        for pydantic_field_name, match_orm_field in matching:
                                            match_value = val
                                            for _field in pydantic_field_name.split('.'):
                                                match_value = getattr(match_value, _field)

                                            if isinstance(match_value, Reference):
                                                match_value = match_value.id

                                            matching_search &= models.Q(**{match_orm_field.field.attname: match_value})

                                        return related_model.objects.filter(**obj_fields).get(matching_search)

                                    elif isinstance(matching, callable):
                                        raise NotImplementedError

                                    else:
                                        raise NotImplementedError

                                else:
                                    raise NotImplementedError

                            except related_model.DoesNotExist:
                                return get_subobj(force_create=True)

                        else:
                            raise NotImplementedError

                    sub_obj = get_subobj()
                    existing_object_ids.discard(sub_obj.id)
                    subobjects.append(sub_obj)
                    sub_transfer = transfer_to_orm(val, sub_obj, exclude_unset=exclude_unset, access=access, action=action, _just_return_objs=True)
                    subobjects += sub_transfer[0]
                    existing_objects += sub_transfer[1]

                existing_objects += related_model.objects.filter(id__in=list(existing_object_ids))

            else:
                raise NotImplementedError

        else:
            raise NotImplementedError

    if subobjects and not action:
        raise AssertionError('action is not defined but subobjects exist')

    if _just_return_objs:
        return subobjects, existing_objects

    if action in (TransferAction.CREATE, TransferAction.SYNC, TransferAction.NO_SUBOBJECTS) and created_submodels is None:
        with atomic():
            should_save: Callable[[models.Model], bool] = lambda obj: bool(not do_not_save_if_no_change or (isinstance(obj, DirtyFieldsMixin) and obj.get_dirty_fields(check_relationship=True)))
            if should_save(django_obj):
                django_obj.save()

            if action in (TransferAction.CREATE, TransferAction.SYNC):
                if action == TransferAction.SYNC:
                    for obj in existing_objects:
                        obj.delete()

                for obj in subobjects:
                    if should_save(obj):
                        obj.save()


async def update_orm(model: Type[BaseModel], orm_obj: models.Model, input: BaseModel, *, access: Optional[Access] = None) -> BaseModel:
    """
    Apply (partial) changes given in `input` to an orm_obj and return an instance of `model` with the full data of the orm including the updated fields.
    """
    warnings.warn("Use transfer_to_orm with exclude_unset=True instead of this function", category=DeprecationWarning)

    if access:
        check_field_access(input, access)

    data = await model.from_orm(orm_obj)
    input_dict: dict = input.dict(exclude_unset=True)

    def update(model: BaseModel, input: dict):
        for key, value in input.items():
            if isinstance(value, dict):
                attr = getattr(model, key)
                if attr is None:
                    setattr(model, key, model.__fields__[key].type_.parse_obj(value))

                else:
                    update(attr, value)

            else:
                setattr(model, key, value)

    update(data, input_dict)

    values, fields_set, validation_error = validate_model(model, data.dict())
    if validation_error:
        raise RequestValidationError(validation_error.raw_errors)

    transfer_to_orm(data, orm_obj)
    return data



from typing import Any, Generator, Mapping, Optional, Tuple, Type, TypeVar, Union
from pydantic import BaseModel, validate_model
from django.db import models
from django.db.models.manager import Manager
from djutils.asyncio import AllowAsyncUnsafe
from ... import context
from ..pydantic import get_orm_field_attr
from .django_to_pydantic import transfer_from_orm

try:
    from fastapi.exceptions import RequestValidationError

except ImportError:
    from pydantic import ValidationError as RequestValidationError


class DjangoORMBaseModel(BaseModel):
    @classmethod
    def from_orm(cls, obj: models.Model, filter_submodel: Optional[Mapping[Manager, models.Q]] = None):
        return transfer_from_orm(cls, obj, filter_submodel=filter_submodel)

    class Config:
        orm_mode = True


def validate_object(obj: BaseModel, is_request: bool = True):
    *_, validation_error = validate_model(obj.__class__, obj.__dict__)
    if validation_error:
        if is_request:
            raise RequestValidationError(validation_error.raw_errors)

        raise validation_error


TDjangoModel = TypeVar('TDjangoModel', bound=models.Model)

def orm_object_validator(model: Type[TDjangoModel], value: Union[str, models.Q]) -> TDjangoModel:
    if isinstance(value, str):
        value = models.Q(id=value)

    access = context.access.get()
    if access and hasattr(model, 'tenant_id'):
        value &= models.Q(tenant_id=access.tenant_id)

    with AllowAsyncUnsafe():
        try:
            return model.objects.get(value)

        except model.DoesNotExist:
            raise ValueError('reference_not_exist')


def get_sync_matching_values(model: BaseModel) -> Generator[Tuple[models.Field, Any], None, None]:
    for name, field in model.__fields__.items():
        if isinstance(field, BaseModel):
            yield from get_sync_matching_values(field)

        if not get_orm_field_attr(field.field_info, 'is_sync_matching_field'):
            continue

        yield (get_orm_field_attr(field.field_info, 'orm_field'), getattr(model, name))


def get_sync_matching_filter(model: BaseModel) -> models.Q:
    fields = {field.field.name: value for field, value in get_sync_matching_values(model)} or {'id': model.id}
    return models.Q(**fields)



import typing
from typing import Callable, ForwardRef, Optional, Type, Any
from types import FunctionType
from pydantic import BaseModel, create_model, Field as Field
from pydantic.fields import ModelField, FieldInfo, SHAPE_SINGLETON, SHAPE_LIST, Undefined
from ..fields import ORMFieldInfo, Field as ORMField


TypingGenericAlias = type(Any)


def _new_field_from_model_field(
    field: ModelField,
    default: Any = Undefined,
    required: Optional[bool] = None
):
    if default is not Undefined:
        default = field.default

    if required is None and field.required and (default is Undefined or field.default is None):
        default = ...

    if isinstance(field.field_info, ORMFieldInfo):
        return ORMField(
            default,
            default_factory=field.default_factory,
            alias=field.alias,
            orm_field=field.field_info.orm_field,
            orm_method=field.field_info.orm_method,
            scopes=field.field_info.scopes,
            is_critical=field.field_info.is_critical,
            sync_matching=field.field_info.sync_matching,
            is_sync_matching_field=field.field_info.is_sync_matching_field,
            **field.field_info.extra,
        )

    return Field(
        default,
        default_factory=field.default_factory,
        alias=field.alias,
        **field.field_info.extra,
    )


class OptionalModel(BaseModel):
    pass


def to_optional(id_key: str = 'id'):
    def wrapped(cls: Type[BaseModel]):
        def optional_model(c, __module__: str, __parent__module__: str):
            try:
                if issubclass(c, BaseModel):
                    field: ModelField
                    fields = {}
                    for key, field in c.__fields__.items():
                        field_type = optional_model(field.outer_type_, __module__=__module__, __parent__module__=__parent__module__)
                        default = field.default
                        if key == id_key and not field.allow_none:
                            default = default or ...

                        elif not field.allow_none:
                            field_type = Optional[field_type]

                        elif field.required:
                            default = default or ...

                        fields[key] = (field_type, _new_field_from_model_field(field, default, required=False))

                    return create_model(
                        f'{c.__qualname__} [O]',
                        __base__=(c, OptionalModel),
                        __module__=c.__module__ if c.__module__ != __parent__module__ else __module__,
                        **fields,
                    )

            except TypeError:
                pass

            return c

        return optional_model(cls, __module__=cls.__module__, __parent__module__=cls.__base__.__module__)

    return wrapped


class ReferencedModel(BaseModel):
    pass


class Reference(BaseModel):
    def __init_subclass__(cls, rel: Optional[str] = None, rel_params: Optional[Callable] = None, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls._rel = getattr(cls, '_rel', rel)
        cls._rel_params = rel_params
        c = cls
        while cls._rel is None:
            if issubclass(c.__base__, Reference):
                c = c.__base__
                if not c:
                    raise AssertionError("Cannot find parent Reference with `rel` set")

                try:
                    cls._rel = c._rel
                    cls._rel_params = c._rel_params

                except AttributeError:
                    pass

            else:
                raise AssertionError


def include_reference(reference_key: str = '$rel', reference_params_key: str = '$rel_params'):
    recreated_models = {}
    def wrapped(cls: Type[BaseModel]):
        def model_with_rel(c: Type, __parent__: Type, __module__: str, __parent__module__: str):
            if isinstance(c, ForwardRef):
                return c, False

            if issubclass(c, BaseModel):
                field: ModelField
                fields = {}
                recreate_model = False
                for key, field in c.__fields__.items():
                    if field.shape not in (SHAPE_SINGLETON, SHAPE_LIST):
                        fields[key] = (field.outer_type_, _new_field_from_model_field(field))
                        continue

                    field_type, recreated_model = model_with_rel(field.type_, c, __module__=__module__, __parent__module__=__parent__module__)
                    if field.type_ != field.outer_type_:
                        field_type = getattr(typing, field.outer_type_._name)[field_type]

                    if field.allow_none:
                        field_type = Optional[field_type]

                    fields[key] = (field_type, _new_field_from_model_field(field))
                    if recreated_model:
                        recreate_model = True

                    try:
                        if issubclass(field_type, Reference):
                            recreate_model = True

                    except TypeError:
                        pass

                if issubclass(c, Reference):
                    recreate_model = True
                    value = Undefined
                    value_example = None
                    value_factory = None
                    if isinstance(c._rel, FunctionType):
                        value_factory = c._rel
                        value_example = c._rel()

                    else:
                        value = value_example = c._rel

                    fields['x_reference_key'] = (str, Field(value, example=value_example, orm_field=None, alias=reference_key, default_factory=value_factory))
                    if c._rel_params:
                        fields['x_reference_params_key'] = (dict, Field(alias=reference_params_key, orm_method=c._rel_params))

                if recreate_model:
                    if c not in recreated_models:
                        recreated_models[c] = create_model(
                            f'{c.__qualname__} [R]',
                            __base__=(c, ReferencedModel),
                            __module__=c.__module__ if c.__module__ != __parent__module__ else __module__,
                            **fields,
                        )
                        recreated_models[c].__recreated__ = True

                    if __parent__:
                        setattr(__parent__, c.__name__, recreated_models[c])

                    return recreated_models[c], True

            return c, False

        return model_with_rel(cls, None, __module__=cls.__module__, __parent__module__=cls.__base__.__module__)[0]

    return wrapped


def is_orm_field_set(field: FieldInfo) -> bool:
    if isinstance(field, ORMFieldInfo):
        orm_field = field.orm_field
        if orm_field is Undefined:
            return False

    else:
        orm_field = field.extra.get('orm_field')
        if 'orm_field' in field.extra and field.extra['orm_field'] is None:
            # Do not raise error when orm_field was explicitly set to None
            return False

    return True


def get_orm_field_attr(field: FieldInfo, key: str):
    if isinstance(field, ORMFieldInfo):
        return getattr(field, key)

    return field.extra.get(key)
